from cronex import CronExpression
from flask import abort, request, redirect, url_for, flash, jsonify
from flask_admin.menu import MenuLink
from wtforms import validators
from wtforms.fields import PasswordField
from wtforms.fields.html5 import EmailField
from flask_admin.form import fields
from flask_admin.model.template import EndpointLinkRowAction
from flask_admin.contrib.sqla import ModelView
from flask_admin import AdminIndexView, expose
from flask_security import login_required, current_user, utils, logout_user, login_user

from skidward.models import db, Task, Job, Message
from skidward.backend import get_redis_backend
from skidward.web.forms import TaskContextForm
from skidward.task_runner import TaskRunner


class SkidwardModelView(ModelView):
    def is_accessible(self):
        return current_user.is_authenticated

    def inaccessible_callback(self, name, **kwargs):
        return redirect(url_for("index.login"))


class UserAdmin(SkidwardModelView):
    form_excluded_columns = ("password",)
    column_exclude_list = ("password",)

    def is_accessible(self):
        return current_user.is_authenticated and current_user.has_role("admin")

    def scaffold_form(self):
        form_class = super(UserAdmin, self).scaffold_form()
        form_class.email = EmailField(
            "Email ID", [validators.DataRequired(), validators.Email()]
        )
        form_class.new_password = PasswordField("New Password")
        form_class.confirm_password = PasswordField(
            "Confirm Password",
            [validators.EqualTo("new_password", message="Passwords do not match")],
        )
        return form_class

    def on_model_change(self, form, model, is_created):
        if len(model.new_password):
            model.password = utils.hash_password(model.new_password)


class RoleAdmin(SkidwardModelView):
    def is_accessible(self):
        return current_user.is_authenticated and current_user.has_role("admin")


class WorkerView(SkidwardModelView):
    can_create = False


class TaskView(SkidwardModelView):

    column_extra_row_actions = [
        EndpointLinkRowAction(
            icon_class="glyphicon glyphicon-cog",
            endpoint="add_context.add_context",
            title="Configure",
            id_arg="task_id",
        ),
        EndpointLinkRowAction(
            icon_class="glyphicon glyphicon-play-circle",
            endpoint="add_context.add_context",
            title="Configured_Run",
            id_arg="task_id",
            url_args=dict(temp_run=True),
        ),
        EndpointLinkRowAction(
            icon_class="glyphicon glyphicon-play",
            endpoint="run_task.run_task",
            title="Quick_Run",
            id_arg="task_id",
        ),
    ]

    def update_model(self, form, model):
        try:
            form.populate_obj(model)
            self._on_model_change(form, model, False)
            self.session.commit()
        except ValueError as ex:
            flash("{}".format(ex), category="warning")
            self.session.rollback()
            return False

        self.after_model_change(form, model, False)

        return True

    def on_model_change(self, form, model, is_created):
        if model.enabled:
            try:
                CronExpression(model.cron_string)
            except (ValueError, AttributeError):
                raise ValueError("Cron String not valid")
            if not model.context:
                raise ValueError("Context needed to enable a task")

    form_overrides = {"context": fields.JSONField}
    form_excluded_columns = ("context",)
    form_create_rules = ("name", "cron_string", "worker")


class JobView(SkidwardModelView):

    column_list = ["task", "ran_at", "state"]
    column_labels = dict(task="Jobs", state="Status")
    can_edit = False

    column_extra_row_actions = [
        EndpointLinkRowAction(
            icon_class="glyphicon glyphicon-eye-open",
            endpoint="view_rendered_logs.view_rendered_logs",
            title="See logs",
            id_arg="job_id",
        )
    ]
    can_edit = False


class SkidwardView(AdminIndexView):
    def is_accessible(self):
        return current_user.is_authenticated

    def inaccessible_callback(self, name, **kwargs):
        return redirect(url_for("index.login"))

    def is_visible(self):
        return False

    @login_required
    @expose("/")
    def index(self):
        return self.render("admin/index.html")

    @expose("/login")
    def login(self):
        login_user(current_user)
        return redirect(url_for("admin.index"))

    @expose("/logout")
    def logout(self):
        logout_user()
        return redirect(url_for("index.login"))


class RunView(SkidwardView):
    @expose("/run/<task_id>")
    @login_required
    def run_task(self, task_id):
        task = Task.query.get(task_id)
        context = request.args.get("context")

        if not context and not task.context:
            flash("Cannot Run Task without Configuration.", category="error")
            return redirect(url_for("add_context.add_context", task_id=task_id))

        redis_client = get_redis_backend()
        if context:
            temp_task = {"task_id": task_id, "overwrite_context": context}
            redis_client.hmset("CONFIGURED_RUN", temp_task)
            flash("Job is Running... Refresh the page to view updated Status")
            return redirect("/admin/job")

        else:
            redis_client.lpush("MANUAL_RUN", task_id)
            flash("Job is Running... Refresh the page to view updated Status")
            return redirect("/admin/job")


class TaskConfigure(SkidwardView):
    @expose("/configure/<task_id>", methods=["GET", "POST"])
    @login_required
    def add_context(self, task_id):
        form = TaskContextForm()
        task = Task.query.get(task_id)

        if task.context:
            form.context.data = task.context

        if form.validate_on_submit():
            dict_of_fieldlist = request.form.to_dict(flat=False)

            keys = dict_of_fieldlist.get("key")
            values = dict_of_fieldlist.get("value")
            context = dict(zip(keys, values)) if keys and values else None

            if request.args.get("temp_run"):
                return redirect(
                    url_for("run_task.run_task", task_id=task.id, context=context)
                )

            else:
                task.context = context
                db.session.commit()
                return redirect(url_for("add_context.index"))

        return self.render("task/configure.html", form=form, name=task.name)


class LiveLogView(SkidwardView):
    @expose("/get_logs/<int:job_id>", methods=["GET"])
    @login_required
    def get_logs(self, job_id):
        job = Job.query.get(job_id)
        logging_key = job.task.context.get("LOGGING_KEY") if job.task.context else None

        if not logging_key:
            logging_key = TaskRunner.get_unique_message_key(job_id, job.task_id)

        messages = get_redis_backend().lrange(logging_key, 0, -1)
        # Assume key no longer exists because messages have been persisted.
        if len(messages) == 0:
            messages = db.session.query(Message).filter(Message.job_id == job_id).all()
            message_content = [m.content for m in messages]

            return jsonify(message_content)

        return jsonify(messages)

    @expose("/logs/<int:job_id>", methods=["GET"])
    @login_required
    def view_rendered_logs(self, job_id):
        return self.render("log/logview.html", job_id=job_id)


class LoginMenuLink(MenuLink):
    def is_accessible(self):
        return not current_user.is_authenticated


class LogoutMenuLink(MenuLink):
    def is_accessible(self):
        return current_user.is_authenticated



import os
import logging
from functools import partialmethod

from redis import StrictRedis

logger = logging.getLogger(__name__)


class RedisProxy:
    def __init__(self, backend):
        self.backend = backend

    def _decode_response(self, values):
        if isinstance(values, bytes):
            return values.decode("utf-8")

        if isinstance(values, list):
            new_values = []
            for element in values:
                if isinstance(element, bytes):
                    element = element.decode("utf-8")
                new_values.append(element)
            return new_values

        return values

    def proxy(self, method, *args, **kwargs):
        errors = {}

        backend_name = self.backend.__class__.__name__
        try:
            logger.debug(
                "trying to {} {} using {}".format(method, kwargs, backend_name)
            )
            bound_method = getattr(self.backend, method)
            result = bound_method(*args, **kwargs)

            return self._decode_response(result)
        except Exception as e:
            if os.getenv("DEBUG"):
                raise
            errors[backend_name] = str(e)

        raise Exception(
            "No backend managed to {} for the provided request: {}".format(
                method, errors
            )
        )

    lpush = partialmethod(proxy, "lpush")
    lrange = partialmethod(proxy, "lrange")
    erase = partialmethod(proxy, "erase")
    set = partialmethod(proxy, "set")
    get = partialmethod(proxy, "get")
    rpop = partialmethod(proxy, "rpop")
    exists = partialmethod(proxy, "exists")
    hset = partialmethod(proxy, "hset")
    expire = partialmethod(proxy, "expire")
    hmset = partialmethod(proxy, "hmset")
    hmget = partialmethod(proxy, "hmget")
    hdel = partialmethod(proxy, "hdel")


class RedisDummyBackend:
    def __init__(self, **kwargs):
        self.redis_lists = {"set": {}}

    def is_task_new(self, name):
        return name not in self.redis_lists and name not in self.redis_lists["set"]

    def lpush(self, name, elements):
        if not isinstance(elements, list):
            elements = [elements]

        elements = [str(element).encode("utf-8") for element in elements]

        if not name in self.redis_lists:
            self.redis_lists[name] = []

        self.redis_lists[name] = elements + self.redis_lists[name]

    def lrange(self, name, start, end):
        if self.is_task_new(name):
            self.redis_lists[name] = []

        if end == -1:
            return self.redis_lists[name][start:]
        else:
            end = end + 1

        result = self.redis_lists[name][start:end]

        return result

    def erase(self):
        self.redis_lists = {}

    def set(self, name, value, ex=None, nx=None):
        if nx:
            if self.is_task_new(name):
                self.redis_lists["set"][name] = str(value).encode("utf-8")
        else:
            self.redis_lists["set"][name] = str(value).encode("utf-8")

    def get(self, name):
        if not self.is_task_new(name):
            return self.redis_lists["set"][name]

    def rpop(self, name):
        if not self.is_task_new(name):
            if self.redis_lists[name]:
                return self.redis_lists[name].pop(-1)

    def exists(self, name):
        return not self.is_task_new(name)

    def hmset(self, name, mapping):
        if not name in self.redis_lists:
            self.redis_lists[name] = []
        self.redis_lists[name].append(mapping)

    def hmget(self, name, keys):
        if not self.is_task_new(name):
            if self.redis_lists[name]:
                task = self.redis_lists[name][-1]
                task_id = task[keys[0]]
                context = task[keys[1]]
                return task_id, context

    def hdel(self, name, *keys):
        if self.redis_lists[name]:
            self.redis_lists[name].pop(-1)

    def expire(self, name, *args):
        if self.redis_lists[name]:
            del self.redis_lists[name]


def get_backend_configuration():
    return {
        "host": os.getenv("REDIS_HOST") or "localhost",
        "port": os.getenv("REDIS_PORT") or 6379,
        "password": os.getenv("REDIS_PASSWORD") or None,
        "db": os.getenv("REDIS_DB") or "0",
    }


def get_redis_backend(**kwargs):
    def _is_testing():
        return os.getenv("WEB_TESTING") is True or os.getenv("WEB_TESTING") in [
            "True",
            "true",
            "TRUE",
        ]

    if _is_testing():
        return RedisProxy(RedisDummyBackend(**kwargs))

    local_config = get_backend_configuration()
    merged_configuration = {**local_config, **kwargs}

    return RedisProxy(StrictRedis(**merged_configuration))



from flask_admin.form import JSONField
from wtforms import StringField, PasswordField
from wtforms.validators import DataRequired, Email
from wtforms.fields.html5 import EmailField
from flask_wtf import FlaskForm
from flask_security.forms import LoginForm
from wtforms.compat import string_types


class KeyValueValidator(DataRequired):
    def __call__(self, form, field):
        if (
            not field.data
            or isinstance(field.data, string_types)
            and not field.data.strip()
        ):
            field.errors[:] = []


class TaskContextForm(FlaskForm):
    key = StringField("Key", [KeyValueValidator()])
    value = StringField("Value", [KeyValueValidator()])
    context = JSONField("Context")


class SkidwardLoginForm(LoginForm):
    email = EmailField("Email ID", validators=[DataRequired(), Email()])
    password = PasswordField("Password", validators=[DataRequired()])



import copy
import os
import time
from itertools import chain
from datetime import datetime as dt
from ast import literal_eval
import multiprocessing as mp

from cronex import CronExpression

from skidward.backend import get_redis_backend
from skidward.models import Task
from skidward.task_runner import TaskRunner


REDIS_ELECTION_TTL = 60
REDIS_MANUAL_CHANNEL = "MANUAL_RUN"
REDIS_CONFIGURED_RUN_CHANNEL = "CONFIGURED_RUN"
WAKEUP_TIMER = int(os.getenv("WAKEUP_TIMER"))

mp_context = mp.get_context("spawn")
process_id = os.getpid()


class Elector:
    def __init__(self, redis_client):
        self.redis_client = redis_client

    def get_tasks(self):
        for task in chain(
            self.get_scheduled_tasks(),
            self.get_manual_tasks(),
            self.get_configured_tasks(),
        ):
            yield task

    def get_scheduled_tasks(self):
        for task in self.get_elected_tasks(Task.query.filter(Task.enabled).all()):
            yield task

    def get_manual_tasks(self):
        task_id = self.redis_client.rpop(REDIS_MANUAL_CHANNEL)
        while task_id:
            yield Task.query.get(task_id)
            task_id = self.redis_client.rpop(REDIS_MANUAL_CHANNEL)

    def get_configured_tasks(self):
        task_id, context = self.redis_client.hmget(
            REDIS_CONFIGURED_RUN_CHANNEL, keys=["task_id", "overwrite_context"]
        )
        while task_id:
            self.redis_client.hdel(
                REDIS_CONFIGURED_RUN_CHANNEL, "task_id", "overwrite_context"
            )
            task = Task.query.get(task_id)
            temp_task = copy.copy(task)
            temp_task.context = literal_eval(context)
            yield temp_task
            task_id, context = self.redis_client.hmget(
                REDIS_CONFIGURED_RUN_CHANNEL, keys=["task_id", "overwrite_context"]
            )

    def get_elected_tasks(self, tasks):
        return [task for task in tasks if self.is_runner_elected_for_task(task)]

    def is_runner_elected_for_task(self, task):
        date_tuple = dt.utcnow().timetuple()[:5]
        date_mask = "-".join(map(str, date_tuple))
        unique_key = f"{task.id}-{date_mask}"

        cron_expression = CronExpression(task.cron_string)
        if cron_expression.check_trigger(date_tuple):
            if not self.redis_client.exists(unique_key):
                self.redis_client.set(
                    unique_key, process_id, ex=REDIS_ELECTION_TTL, nx=True
                )
                return self.redis_client.get(unique_key) == str(process_id)

        return False


class SchedulerRunner:
    def __init__(self):
        self.elector = Elector(get_redis_backend())

    def run(self):
        while True:
            for task in self.elector.get_tasks():
                p = mp_context.Process(
                    target=TaskRunner, args=(task.id, task.context, task.worker.name)
                )
                p.start()

            time.sleep(WAKEUP_TIMER)


def start():
    SchedulerRunner().run()



import logging
import datetime
from unittest import mock

from skidward.app import app
from skidward.models import Job, JobStatus, Message, Task, Worker
from skidward.task_runner import TaskRunner


@mock.patch("skidward.task_runner.TaskRunner.start")
def test_new_job_is_created(mock_start, init_database, monkeypatch):
    monkeypatch.setattr(
        TaskRunner,
        "_set_logging_config",
        (lambda x: logging.basicConfig(handlers=logging.NullHandler())),
    )
    with app.app_context():
        task = Task.query.one()
        task_runner = TaskRunner(task.id, task.context, task.worker.name)
        task_runner._create_new_job()

        assert init_database.session.query(Job).one().task.id == task.id


def test_unique_message_key_creation(init_database):
    test_worker = Worker(name="real_name")
    test_task = Task(
        name="task_1",
        worker=test_worker,
        context={"setting_1": "value_1"},
        cron_string="*/2 * * * *",
    )
    test_job = Job(task=test_task)
    init_database.session.commit()

    assert TaskRunner.get_unique_message_key(
        test_job.id, test_task.id
    ) == "J{}-T{}".format(test_job.id, test_task.id)


@mock.patch("skidward.worker_detector.create_namespace_module_manager")
def test_full_process_calling_run(mock_create, init_database, monkeypatch):
    monkeypatch.setattr(
        TaskRunner,
        "_set_logging_config",
        (lambda x, y: logging.basicConfig(handlers=logging.NullHandler())),
    )
    monkeypatch.setattr(
        TaskRunner,
        "_persist_logs",
        (lambda x, y, z: print("Things got saved to the DB.")),
    )
    monkeypatch.setattr(
        TaskRunner, "_expire_logs", (lambda x, y: print("Original list got expired."))
    )
    with app.app_context():
        task = Task.query.all()[0]
        context = {"LOGGING_KEY": "ch:channel"}
        task.context = context
        init_database.session.commit()

        TaskRunner(task.id, task.context, task.worker.name)
        job = init_database.session.query(Job).one()

        mock_create.return_value.load_module.assert_called_once_with("real_name")
        mock_create.return_value.load_module.return_value.start.assert_called_once_with(
            {"LOGGING_KEY": "ch:channel"}
        )
        assert job.id == 1
        assert job.task_id == 1
        assert job.state == JobStatus.SUCCESS
        assert isinstance(job.ran_at, datetime.datetime)


@mock.patch("skidward.task_runner.TaskRunner.start")
def test_redis_list_name_for_log_is_same_as_given(mock_start, init_database):
    task = Task.query.one()
    job = Job(task=task)
    init_database.session.commit()

    task.context = {"LOGGING_KEY": "TEST"}

    runner = TaskRunner(task.id, task.context, task.worker.name)
    logging_key = runner._get_logging_key(job)
    assert logging_key == "TEST"


@mock.patch("skidward.task_runner.TaskRunner.start")
def test_redis_list_name_for_log_is_automatically_build(mock_start, init_database):
    task = Task.query.one()
    job = Job(task=task)
    init_database.session.add(job)
    init_database.session.commit()

    job = Job.query.one()
    runner = TaskRunner(task.id, task.context, task.worker.name)
    logging_key = runner._get_logging_key(job)
    assert logging_key == "J{}-T{}".format(job.id, task.id)


@mock.patch("skidward.task_runner.TaskRunner.start")
def test_set_expire_on_redis_list_delete_it(mock_start, backend, init_database):
    task = Task.query.one()
    runner = TaskRunner(task.id, task.context, task.worker.name)
    runner.redis_client = backend
    backend.lpush("logging_key", "TEST")
    assert backend.lrange("logging_key", 0, -1)
    runner._expire_logs("logging_key")
    assert not backend.lrange("logging_key", 0, -1)


@mock.patch("skidward.task_runner.TaskRunner.start")
def test_set_expire_on_redis_list_delete_it(mock_start, backend, init_database):
    with app.app_context():
        task = Task.query.one()
        job = Job(task=task)
        init_database.session.add(job)
        init_database.session.commit()
        job = Job.query.one()

        runner = TaskRunner(task.id, task.context, task.worker.name)
        runner.redis_client = backend

        logs = ["log_one", "log_two", "log_three"]
        backend.lpush("logging_key", logs)

        runner._persist_logs(job, "logging_key")
        messages = Message.query.filter(Job.id == job.id).all()

        for message, log in zip(messages, logs):
            assert message.content == log



from unittest import mock

from flask_login import current_user

from skidward.models import Job, Message, Task, User, Worker


def test_it_loads_the_app(test_client, pass_security_context_processor):
    response = test_client.get("/login")
    assert response.status_code == 200
    assert b"Login" in response.data


def test_it_requires_login_to_access_home_page(
    test_client, pass_security_context_processor
):
    with test_client:
        response = test_client.get("/", follow_redirects=True)
        assert b"Login" in response.data


def test_it_logs_in_correct_user(test_client, init_database):
    with test_client:
        response = test_client.post(
            "/login",
            data=dict(email="test@test.com", password="123"),
            follow_redirects=True,
        )
        assert current_user.email == "test@test.com"


def test_it_creates_new_user_in_the_database(test_client, init_database):
    user = User.query.filter_by(email="test@test.com").first()
    assert user is not None
    assert user.email == "test@test.com"
    assert user.username == "testuser"


def test_it_redirects_to_admin_interface_on_login(test_client, init_database):
    response = test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    assert b"Logged in as" in response.data


def test_it_logs_out_correctly_and_redirects_to_login(
    test_client, pass_security_context_processor
):
    with test_client:
        response = test_client.get("/logout", follow_redirects=True)
        assert response.status_code == 200
        assert b"Login" in response.data


def test_configure_asks_for_context_and_save_in_db_correctly(
    test_client, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    response = test_client.get("/admin/task/configure/1", follow_redirects=True)
    assert b"Add Configuration" in response.data
    test_client.post(
        "/admin/task/configure/1",
        data=dict(key="new_key", value="new_value"),
        follow_redirects=True,
    )
    task = Task.query.get(1)
    assert task.context == {"new_key": "new_value"}
    assert type(task.context) == dict


@mock.patch("skidward.backend.RedisDummyBackend")
def test_configured_run_takes_new_context_but_does_not_save_in_db(
    mock_redis, backend, test_client, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    mock_redis.return_value = backend
    response = test_client.post(
        "/admin/task/configure/1?temp_run=True",
        data=dict(key="tmp_key", value="tmp_value"),
        follow_redirects=True,
    )
    task_id, context = backend.hmget(
        "CONFIGURED_RUN", keys=["task_id", "overwrite_context"]
    )
    assert context == "{'tmp_key': 'tmp_value'}"
    task = Task.query.get(1)
    assert task.context == {"test_key": "test_value"}


@mock.patch("skidward.backend.RedisDummyBackend")
def test_quick_run_pushes_task_id_to_redis(
    mock_redis, test_client, backend, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    mock_redis.return_value = backend
    response = test_client.get("/admin/task/run/1", follow_redirects=True)
    redis_tasks = backend.lrange("MANUAL_RUN", 0, -1)
    assert response.status_code == 200
    assert "1" in redis_tasks


@mock.patch("skidward.backend.RedisDummyBackend")
def test_configured_run_pushes_dict_of_task_with_context_to_redis(
    mock_redis, test_client, backend, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    mock_redis.return_value = backend
    test_client.post(
        "admin/task/configure/1?temp_run=True",
        data=dict(key="tmp_key", value="tmp_value"),
        follow_redirects=True,
    )
    task_id, context = backend.hmget(
        "CONFIGURED_RUN", keys=["task_id", "overwrite_context"]
    )
    assert context == "{'tmp_key': 'tmp_value'}"


def test_manual_run_redirects_to_configure_if_no_context_is_present(
    test_client, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    worker = Worker.query.first()
    new_task = Task(name="New_Task", worker=worker)
    init_database.session.add(new_task)
    init_database.session.commit()
    task = Task.query.filter_by(name="New_Task").first()
    assert task is not None
    assert task.id == 2
    assert not task.context
    response = test_client.get("/admin/task/run/2", follow_redirects=True)
    assert b"Add Configuration" in response.data


def test_end_user_has_no_access_to_user_view_and_get_redirected(
    test_client, init_database
):
    response = test_client.get("admin/user/")

    assert response.status_code == 302
    assert response.location == "http://localhost/login"


def test_admin_has_access_to_user_view(test_admin_client, init_database):
    response = test_admin_client.get("admin/user/")

    assert response.status_code == 200


def test_to_get_logs_from_db_when_no_longer_in_redis(
    test_admin_client, init_database, backend
):
    logged_message = "Logging job test"
    task = Task.query.one()
    job = Job(task_id=task.id)
    init_database.session.add(job)
    init_database.session.commit()

    message = Message(job_id=job.id, content=logged_message)
    init_database.session.add(message)
    init_database.session.commit()

    response = test_admin_client.get("admin/logs/get_logs/{}".format(job.id))

    assert response.status_code == 200
    assert response.json == [logged_message]


@mock.patch("skidward.web.views.get_redis_backend")
def test_to_get_logs_from_redis(mock_redis, test_admin_client, init_database, backend):
    mock_redis.return_value = backend

    logged_message = "Logging job test"
    task = Task.query.one()
    task.context = {"LOGGING_KEY": "TEST"}
    job = Job(task_id=task.id)
    init_database.session.add(job)
    init_database.session.commit()

    backend.lpush("TEST", logged_message)

    response = test_admin_client.get("admin/logs/get_logs/{}".format(job.id))

    assert response.status_code == 200
    assert response.json == [logged_message]


def test_calling_logs_endpoint(test_admin_client, init_database):
    task = Task.query.one()
    job = Job(task_id=task.id)
    init_database.session.add(job)
    init_database.session.commit()

    response = test_admin_client.get("admin/logs/logs/{}".format(job.id))

    assert response.status_code == 200
    assert response.content_type == "text/html; charset=utf-8"



import datetime
import logging

from redis_log_handler import RedisKeyHandler

from skidward.app import app
from skidward import worker_detector
from skidward.models import db, Job, JobStatus, Message, Task
from skidward.backend import get_redis_backend, get_backend_configuration


MESSAGE_EXPIRATION_DURATION = 10


class TaskRunner:
    def __init__(self, task_id, context, worker_name):
        with app.app_context():
            self.task = Task.query.get(task_id)
            self.context = context
            self.worker_name = worker_name
            self.redis_client = get_redis_backend()

            self.start()

    @staticmethod
    def get_unique_message_key(job_id, task_id):
        return "J{}-T{}".format(job_id, task_id)

    def _create_new_job(self):
        job = Job(
            state=JobStatus.RUNNING,
            ran_at=datetime.datetime.utcnow(),
            task_id=self.task.id,
        )
        db.session.add(job)
        db.session.commit()

        return job

    def _get_logging_key(self, job):
        key = self.context.get("LOGGING_KEY")
        if key:
            return key

        return TaskRunner.get_unique_message_key(job.id, self.task.id)

    def _set_logging_config(self, logging_key):
        raw_logs = self.context.get("RAW_LOGS", False)
        log_handler = RedisKeyHandler(
            logging_key, raw_logging=raw_logs, **get_backend_configuration()
        )
        logger = logging.getLogger()
        logger.addHandler(log_handler)
        logger.setLevel(logging.INFO)

        logging.info("Logging {}".format(self.worker_name))

    def _persist_logs(self, job, logging_key):
        def _get_all_log_messages():
            return self.redis_client.lrange(logging_key, 0, -1)

        messages = [
            Message(job_id=job.id, content=msg) for msg in _get_all_log_messages()
        ]
        db.session.bulk_save_objects(messages)

    def _expire_logs(self, logging_key):
        self.redis_client.expire(logging_key, MESSAGE_EXPIRATION_DURATION)

    def start(self):
        # Need to create a job to fallback to a unique message key for Redis
        # happening before the first logging takes place.
        job = self._create_new_job()

        logging_key = self._get_logging_key(job)
        self._set_logging_config(logging_key)

        logging.info("Job id:{} created".format(job.id))

        worker_module = worker_detector.load_worker_on_namespace(self.worker_name)

        logging.info("{} is running".format(self.worker_name))

        try:
            worker_module.start(self.context)
            status = JobStatus.SUCCESS
        except Exception as e:
            status = JobStatus.FAIL
            logging.info(e)

        logging.info("Status : {}".format(status))

        job.state = status
        self._persist_logs(job, logging_key)
        db.session.commit()

        self._expire_logs(logging_key)



import os

from unittest import mock
import pytest
import redis

from skidward.backend import get_redis_backend, RedisProxy, RedisDummyBackend


class TestRedisBackend:
    def test_get_expected_backend(self):
        backend = get_redis_backend()
        assert backend.__class__.__name__ == "RedisProxy"
        assert backend.backend.__class__.__name__ == "RedisDummyBackend"

    @mock.patch("skidward.backend.StrictRedis")
    def test_get_actual_redis_backend(self, mock_redis):
        # TESTING env var is added via the autouse "enable_testing" fixture
        # Need to remove it to check the real case behavior
        os.environ.pop("WEB_TESTING")
        backend = get_redis_backend()
        assert backend.__class__.__name__ == "RedisProxy"
        assert backend.backend.__class__.__name__ == "MagicMock"

    @pytest.mark.skip(
        reason="What is the benefit of this test? It's frail and seems to test a mock implementation."
    )
    @mock.patch("skidward.backend.StrictRedis")
    def test_call_actual_redis_commands(self, mock_redis):
        backend = get_redis_backend()
        backend.lpush("jobs", 1)
        backend.lrange("jobs", 0, -1)

        calls = [
            mock.call(),
            mock.call().lpush("jobs", 1),
            mock.call().lrange("jobs", 0, -1),
        ]

        assert mock_redis.mock_calls == calls


class TestRedisProxy:
    @pytest.mark.parametrize("expected_class", (RedisDummyBackend, redis.Redis))
    def test_proxy_returns_expected_class(self, expected_class):
        backend = RedisProxy(expected_class)
        assert backend.backend == expected_class

    def test_do_not_compute_string_just_return_it(self, backend):
        input_expected = "task_1"
        assert backend._decode_response(input_expected) == input_expected

    def test_do_compute_byte_to_string(self, backend):
        input_expected = "task_id"
        assert (
            backend._decode_response(input_expected.encode("utf-8")) == input_expected
        )

    def test_do_compute_list_containing_bytes_to_list_of_string(self, backend):
        input = ["task_1", "task_2".encode("utf-8")]
        expected = ["task_1", "task_2"]
        assert backend._decode_response(input) == expected


class TestRedisDummyBackend:
    @pytest.mark.parametrize("list_name, expected", (("job", True), ("nope", False)))
    def test_is_task_new_in_redis_list(self, list_name, expected):
        dummy_backend = RedisDummyBackend()
        assert dummy_backend.is_task_new(list_name)
        dummy_backend.redis_lists["job"] = [1]
        assert not dummy_backend.is_task_new(list_name) is expected

    @pytest.mark.parametrize("jobs_id", (2, [2]))
    def test_lpush_handles_int_and_list(self, jobs_id):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["job"] = ["1".encode("utf-8")]

        dummy_backend.lpush("job", jobs_id)

        assert dummy_backend.redis_lists["job"] == [
            "2".encode("utf-8"),
            "1".encode("utf-8"),
        ]

    @pytest.mark.parametrize("list_name", ("job", "add_me"))
    def test_lpush_creates_redis_list_if_not_existing(self, list_name):
        dummy_backend = RedisDummyBackend()
        assert len(dummy_backend.redis_lists.keys()) == 1

        dummy_backend.lpush(list_name, 1)

        assert len(dummy_backend.redis_lists.keys()) == 2
        jobs = dummy_backend.redis_lists.keys()
        assert list_name in jobs

    @pytest.mark.parametrize("values", ([1], [1, 2], [1, 2, 3]))
    def test_lrange_on_an_existing_job_returns_values(self, values):
        dummy_backend = RedisDummyBackend()
        assert not dummy_backend.lrange("job", 0, -1)
        dummy_backend.redis_lists["job"] = values

        result = dummy_backend.lrange("job", 0, -1)

        assert values == result

    @pytest.mark.parametrize(
        "start, end, expected",
        ((0, -1, [1, 2, 3, 4, 5]), (0, 3, [1, 2, 3, 4]), (1, 3, [2, 3, 4])),
    )
    def test_lrange_returns_expected_segment(self, start, end, expected):
        dummy_backend = RedisDummyBackend()
        assert not dummy_backend.lrange("job", start, end)
        dummy_backend.redis_lists["job"] = [1, 2, 3, 4, 5]

        result = dummy_backend.lrange("job", start, end)

        assert result == expected

    @pytest.mark.parametrize(
        "start, end, expected",
        (
            (-10, 0, [1]),
            (-10, 10, [1, 2, 3, 4, 5]),
            (0, -10, []),
            (6, 10, []),
            (4, 2, []),
        ),
    )
    def test_lrange_when_start_end_are_out_of_bounds(self, start, end, expected):
        dummy_backend = RedisDummyBackend()
        assert not dummy_backend.lrange("job", start, end)
        dummy_backend.redis_lists["job"] = [1, 2, 3, 4, 5]

        result = dummy_backend.lrange("job", start, end)

        assert result == expected

    def test_erase_deletes_redis_list(self):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["job"] = [1, 2, 3, 4, 5]
        assert dummy_backend.redis_lists
        dummy_backend.erase()
        assert not dummy_backend.redis_lists

    def test_exception(self):
        dummy_backend = RedisDummyBackend()
        with pytest.raises(AttributeError) as attr_error:
            dummy_backend.fake_function()
        assert "'RedisDummyBackend' object has no attribute 'fake_function'" == str(
            attr_error.value
        )

    @pytest.mark.parametrize(
        "exists, expected", ((True, "host_id_1"), (False, "host_id_2"))
    )
    def test_set_creates_list_with_only_one_value(self, exists, expected):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["set"]["task_key_1"] = "host_id_1".encode("utf-8")
        dummy_backend.set("task_key_1", "host_id_2", ex=None, nx=exists)

        assert dummy_backend.redis_lists["set"]["task_key_1"] == expected.encode(
            "utf-8"
        )

    def test_get_function_return_from_set_list(self):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["set"]["task_key_1"] = "host_id_1".encode("utf-8")
        assert "host_id_1".encode("utf-8") == dummy_backend.get("task_key_1")

    def test_rpop_return_last_element_and_remove_it(self):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["task_key_1"] = ["host_1", "host_2"]
        assert len(dummy_backend.redis_lists["task_key_1"]) == 2
        assert "host_2" == dummy_backend.rpop("task_key_1")
        assert len(dummy_backend.redis_lists["task_key_1"]) == 1
        assert dummy_backend.redis_lists["task_key_1"] == ["host_1"]

    @pytest.mark.parametrize(
        "list_name, expected", (("task_key_1", False), ("task_key_2", True))
    )
    def test_exists_function_return_true_only_if_list_exists(self, list_name, expected):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["task_key_1"] = "host_1"
        assert dummy_backend.is_task_new(list_name) == expected

    @pytest.mark.parametrize(
        "list_name, expected", (("task_key_1", False), ("task_key_2", True))
    )
    def test_exists_function_return_true_only_if_set_exists(self, list_name, expected):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["set"]["task_key_1"] = "host_1"
        assert dummy_backend.is_task_new(list_name) == expected

    @pytest.mark.parametrize("list_name", ("job", "add_me"))
    def test_hmset_creates_redis_list_if_not_existing(self, list_name):
        dummy_backend = RedisDummyBackend()
        assert len(dummy_backend.redis_lists.keys()) == 1
        new_dict = {"a": 1, "b": "test_b"}
        dummy_backend.hmset(list_name, new_dict)

        assert len(dummy_backend.redis_lists.keys()) == 2
        jobs = dummy_backend.redis_lists.keys()
        assert list_name in jobs

    @pytest.mark.parametrize(
        "my_dict", ({"id": 1, "name": "dict1"}, {"id": 2, "name": "dict2"})
    )
    def test_hmset_pushes_dict_in_redis(self, my_dict):
        dummy_backend = RedisDummyBackend()
        dummy_backend.hmset("job", my_dict)
        assert my_dict in dummy_backend.redis_lists["job"]

    def test_hmget_returns_the_last_dict_from_redis(self):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["configured_tasks"] = [
            {"id": 1, "name": "dict1"},
            {"id": 2, "name": "dict2"},
        ]
        assert len(dummy_backend.redis_lists["configured_tasks"]) == 2
        assert dummy_backend.hmget("configured_tasks", ["id", "name"]) == (2, "dict2")

    def test_hdel_deletes_the_last_dict_from_redis(self):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["configured_tasks"] = [
            {"id": 1, "name": "dict1"},
            {"id": 2, "name": "dict2"},
        ]
        assert len(dummy_backend.redis_lists["configured_tasks"]) == 2
        dummy_backend.hdel("configured_tasks", ["id", "name"])
        assert dummy_backend.redis_lists["configured_tasks"] == [
            {"id": 1, "name": "dict1"}
        ]

    def test_list_is_deleted_if_expire(self):
        dummy_backend = RedisDummyBackend()
        dummy_backend.redis_lists["job"] = ["1"]

        assert "job" in dummy_backend.redis_lists
        dummy_backend.expire("job")
        assert not "job" in dummy_backend.redis_lists
